pub mod classes;
#[cfg(test)]
pub mod tests;
#[cfg(test)]
pub mod test_suite;

pub trait WeakFunctor {
    type F<'a, A>;
}

pub trait Functor: WeakFunctor {
    fn fmap<'a, A: 'a, B: 'a, F: 'a + FnOnce(A) -> B>(f: F, fa: Self::F<'a, A>) -> Self::F<'a, B>;

    fn replace<'a, A: 'a, B: 'a>(fa: Self::F<'a, A>, b: B) -> Self::F<'a, B> {
        Self::fmap(|_| b, fa)
    }

    fn void<'a, A: 'a>(fa: Self::F<'a, A>) -> Self::F<'a, ()> {
        Self::replace(fa, ())
    }
}

pub trait ApplicativeSeq: Functor {
    fn seq<'a, A: 'a, B: 'a, F: 'a + FnOnce(A) -> B>(
        ff: Self::F<'a, F>,
        fa: Self::F<'a, A>,
    ) -> Self::F<'a, B>;
    fn _la2<'a, A: 'a, B: 'a, C: 'a, F: 'a + FnOnce(A, B) -> C>(
        f: F,
        fa: Self::F<'a, A>,
        fb: Self::F<'a, B>,
    ) -> Self::F<'a, C> {
        Self::seq(Self::fmap(|a| |b| f(a, b), fa), fb)
    }
}

pub trait ApplicativeLA2: Functor {
    fn la2<'a, A: 'a, B: 'a, C: 'a, F: 'a + FnOnce(A, B) -> C>(
        f: F,
        fa: Self::F<'a, A>,
        fb: Self::F<'a, B>,
    ) -> Self::F<'a, C>;
    fn _seq<'a, A: 'a, B: 'a, F: 'a + FnOnce(A) -> B>(
        ff: Self::F<'a, F>,
        fa: Self::F<'a, A>,
    ) -> Self::F<'a, B> {
        Self::la2(|f, a| f(a), ff, fa)
    }
}

pub trait Applicative: Functor + ApplicativeSeq + ApplicativeLA2 {
    fn pure<'a, A: 'a>(a: A) -> Self::F<'a, A>;

    fn discard_first<'a, A: 'a, B: 'a>(fa: Self::F<'a, A>, fb: Self::F<'a, B>) -> Self::F<'a, B> {
        Self::seq(Self::replace(fa, |b| b), fb)
    }

    fn discard_second<'a, A: 'a, B: 'a>(fa: Self::F<'a, A>, fb: Self::F<'a, B>) -> Self::F<'a, A> {
        Self::la2(|a, _| a, fa, fb)
    }
}

pub trait Monad: Applicative {
    fn bind<'a, A: 'a, B: 'a, F: 'a + FnOnce(A) -> Self::F<'a, B>>(
        fa: Self::F<'a, A>,
        f: F,
    ) -> Self::F<'a, B>;

    fn join<'a, A: 'a>(ffa: Self::F<'a, Self::F<'a, A>>) -> Self::F<'a, A>
    where
        Self::F<'a, A>: 'a,
    {
        Self::bind(ffa, |fa| fa)
    }
}

