use std::rc::Rc;

use super::{tests::*, *};

pub trait FunctorTestSuite: WeakFunctor + Eqr {
    fn sample<'a, A>() -> Vec<Box<dyn 'a + Fn(A) -> Self::F<'a, A>>>;
}

pub fn functor_follows_laws<T: Functor + FunctorTestSuite>() -> R {
    let mut res = R::default();
    for pa in T::sample::<_>() {
        res += fmap_respects_identity::<T, _, _>(|| pa(2));
    }
    for pa in T::sample::<_>() {
        res += fmap_respects_composition::<T, _, _, _, _, _, _>(|x| x + 5, |x| x + 3, || pa(2));
    }
    res
}

pub fn applicative_follows_laws<T: Applicative + FunctorTestSuite>() -> R {
    let mut res = functor_follows_laws::<T>();
    for pa in T::sample::<_>() {
        res += seq_respects_identity::<T, _, _>(|| pa(2));
    }
    for pa in T::sample::<_>() {
        for pg in T::sample::<_>() {
            for pf in T::sample::<_>() {
                res += seq_respects_composition::<T, _, _, _, _, _, _, _, _>(
                    || pf(|x| x + 5),
                    || pg(|x| x + 3),
                    || pa(2),
                );
            }
        }
    }
    res += seq_is_homomorphic::<T, _, _, _, _>(|x| x + 3, || 2);
    for pf in T::sample::<_>() {
        res += seq_respects_interchange::<T, _, _, _, _, _>(|| pf(|x| x + 3), || 2);
    }
    for pa in T::sample::<_>() {
        for pf in T::sample::<_>() {
            res += seq_can_be_expressed_via_la2::<T, _, _, _, _, _>(|| pf(|x| x + 3), || pa(2));
        }
    }
    for pa in T::sample::<_>() {
        res += fmap_can_be_expressed_via_seq::<T, _, _, _, _>(|x| x + 3, || pa(2));
    }
    for pa in T::sample::<_>() {
        for pb in T::sample::<_>() {
            res += discard_can_be_expressed_via_seq_or_la2::<T, _, _, _, _>(|| pa(2), || pb(2));
        }
    }
    res
}

pub fn monad_follows_laws<T: Monad + FunctorTestSuite>() -> R
where {
    let mut res = applicative_follows_laws::<T>();
    for pa in T::sample::<_>() {
        res += bind_respects_left_identity::<T, _, _, _, _>(move |x| pa(x + 3), || 2);
    }
    for pa in T::sample::<_>() {
        res += bind_respects_right_identity::<T, _, _>(move || pa(2));
    }
    for pa in T::sample::<_>() {
        for pg in T::sample::<_>() {
            let pgrc = Rc::new(pg);
            for pf in T::sample::<_>() {
                let pgrci = pgrc.clone();
                let pfrc = Rc::new(pf);
                res += bind_is_associative::<T, _, _, _, _, _, _>(
                    move |x| pfrc(x + 5),
                    move |x| pgrci(x + 3),
                    || pa(2),
                );
            }
        }
    }
    for pf in T::sample::<_>() {
        for pa in T::sample::<_>() {
            seq_can_be_expressed_via_bind::<T, _, _, _, _, _>(|| pf(|x| x + 3), move || pa(2));
        }
    }
    for pa in T::sample::<_>() {
        res += fmap_can_be_expressed_via_bind::<T, _, _, _, _>(|x| x + 3, move || pa(2));
    }
    res
}
